#
# Copyright (C) 2016-2017 Mattia Basaglia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import re
from .. import color as color_ns
from .. import palette


_rgb_color_regex = re.compile("^#[0-9a-fA-F]{6}$")


def string_to_color(color):
    if not color:
        return None
    if _rgb_color_regex.match(color):
        return color_ns.RgbColor(
            int(color[1:3], 16),
            int(color[3:5], 16),
            int(color[5:7], 16)
        )
    if color in palette.colors16.names:
        return color_ns.IndexedColor(str(color), palette.colors16)
    if color in palette.colors256.names:
        return color_ns.IndexedColor(str(color), palette.colors256)
    return None
